/*
 **********************************************************************************************************
 *
 *  (c) Polisoft Design, 2010. All rights reserved.
 *
 *  This code is free software; you can redistribute it and/or modify it under the terms of the GNU
 *  Library General Public License as published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This code is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
 *  the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library
 *  General Public License for more details.
 *
 *  File:       $Workfile:  TCAlign.cpp$
 *              $Revision:  0.0.1$
 *              $Date:      2012-01-18$
 *
 *  Summary:    Trailing Comments Formatter
 *
 *  Project:    TCAlign
 *
 *  Authors:    spoli
 *
 *  Descript:   TCAlign is a trailing comment formatting utility for C/C++ source files that has no effect
 *              on C style comment blocks (i.e. multi-line comments).
 *
 *              Command line syntax with explanations:
 *
 *                  TCAlign -fi <input_file> [-fo <output_file>] [options]
 *
 *              where the input file name (fully qualified, including path if necessary) designates a
 *              text file that would be interpreted as C/C++ source file. The output file name is optional:
 *              if missing, the utility will create a backup file before generating the output so that you
 *              have the option to revert to the original. If the output file name is indicated, the program
 *              will use it for output; dont use the same name as input file name or you will get an error.
 *
 *              The options are:
 *
 *                  -cpp    use C++ comments style (default is C style);
 *                  -th     sensitivity threshold for trailing comments (default is 25);
 *                  -cpos   trailing comment preferred position (default is 65);
 *                  -csz    trailing comment preferred size (default is 42);
 *                  -help   this help text (other arguments are ignored).
 *
 *  Notes:      Make sure that 3 < th < cpos; eventually, TCAling will swap th and cpos values in an attempt
 *              to order them correctly; the minimum value for th, cpos and csz is 4; a smaller value will
 *              be quietly replaced with its corresponding default.
 *
 *  Examples:
 *
 *              Align trailing comments and change their style from C to C++ style using the default
 *              parameters:
 *
 *                  tcalign -fi acmgmt.c -fo acmgmt_fmt -cpp
 *
 *              Align trailing comments and change their style from C++ to C style using a threshold column
 *              of 80 and a comment length of 50:
 *
 *                  tcalign -fi acmgmt.c -fo acmgmt_fmt -cpos 80 -csz 50
 *
 *  URL:        http://polisoftdesign.com/2012/01/18/trailing-comments-formatter/
 *
 **********************************************************************************************************
 *  $Log:  $
 **********************************************************************************************************
 */

/* ******************************************************************************************************** */
/* Include Statements                                                                                       */
/* ******************************************************************************************************** */

#include "stdafx.h"
#include <stdio.h>
#include <string.h>
#include <process.h>
#include <ctype.h>
#include <stdlib.h>
#include <errno.h>

/* ******************************************************************************************************** */
/* Define Constants                                                                                         */
/* ******************************************************************************************************** */

#define RET_OK                  0
#define ERROR_NO_ARGS           1
#define ERROR_NO_INPUT          2
#define ERROR_NO_OUTPUT         3
#define ERROR_SAME_FILE         4
#define ERROR_FILE_NOT_FOUND    5
#define ERROR_CANNOT_CREATE     6
#define ERROR_RENAME            7
#define ERROR_DELETE            8

#define QUOTE                   0x27
#define DOUBLE_QUOTE            0x22

#define MAX_FN_LENGTH           1024
#define MAX_LINE_LENGTH         4096

#define DEFAULT_TSH             25
#define DEFAULT_CM_START        65
#define DEFAULT_CM_SIZE         42
#define MIN_COL_TSH             4

#define LINE_OK                 0
#define PRE_COMMENT             1
#define C_COMMENT               2
#define CPP_COMMENT             3
#define POST_COMMENT            4
#define IN_LITERAL              5
#define NOT_TRL_COMMENT         6

#define CM_ERROR                7

/* ******************************************************************************************************** */
/* Local Variables                                                                                          */
/* ******************************************************************************************************** */

char                           *sz_state[] = {"OK    ",
                                              "PRE_CM",
                                              "C___CM",
                                              "CPP_CM",
                                              "PST_CM",
                                              "IN_LIT",
                                              "NO_TRL",
                                              "?ERROR",
                                              "--09--",
                                              "--10--",
                                              "--11--",
                                              "--12--",
                                              "--13--",
                                              NULL};

char                           *help_text =
                                "TCAlign is a trailing comment formatter for C/C++ source files\n"
                                "It has no effect on C style blocks of comments and on comments\n"
                                "starting in a position before the sensitivity threshold.\n\n"
                                "Usage (arguments in any order):\n\n"
                                "    TCAlign -fi <input_file> [-fo <output_file>] [options]\n\n"
                                "where the options may be:\n\n"
                                "    -cpp    use C++ comments style (default is C style)\n"
                                "    -th     sensitivity threshold for trailing comments (default is 25)\n"
                                "    -cpos   trailing comment preferred position (default is 65)\n"
                                "    -csz    trailing comment preferred size (default is 42)\n"
                                "    -help   this help text (other arguments are ignored)\n\n"
                                "Notes:\n\n"
                                " 1. please make sure that 3 < th < cpos; eventually, TCAlign will\n"
                                "    swap th and cpos values in an attept to order them correctly.\n"
                                " 2. the minimum value for th, cpos and csz is 4; a smaller value\n"
                                "    will be quietly replaced with its corresponding default.\n";

char                           *samp_text =
                                "Test comment in double quotes /* 1234 */ with C   style\n"
                                "Test comment in double quotes // 1234 // with C++ style\n"
                                "strange break between lines\
                                 n now on new line\n";


/* ******************************************************************************************************** */
/* Implementation                                                                                           */
/* ******************************************************************************************************** */

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
int                             line_analyze                    (char  *line_in,
                                                                 int    istate,
                                                                 int   *pcstart,
                                                                 int   *pcend)
{/////////////////////////////////////////////////////////////////////////////////////////////////////////////
    int                         state       = istate;
    int                         i_start     = 0;
    int                         i_end       = 0;
    int                         non_ws      = 0;
    int                         i;

    if ((line_in == NULL) ||
        (pcstart == NULL) ||
        (pcend == NULL))
    {
        return CM_ERROR;
    }

    for (i = 0; line_in[i]; i++)
    {
        switch (state)
        {
        case LINE_OK:
            if (line_in[i] == '/')
            {
                state = PRE_COMMENT;
            }
            else if ((line_in[i] == QUOTE)   ||
                     (line_in[i] == DOUBLE_QUOTE))
            {
                state = IN_LITERAL;
            }
            else if (!isspace (line_in[i]))
            {
                non_ws++;
            }
            break;

        case IN_LITERAL:
            if ((line_in[i] == QUOTE)   ||
                (line_in[i] == DOUBLE_QUOTE))
            {
                state = LINE_OK;
                non_ws++;
            }
            break;

        case PRE_COMMENT:
            if (line_in[i] == '/')
            {
                i_start = i + 1;
                state = CPP_COMMENT;
            }
            else if (line_in[i] == '*')
            {
                state = C_COMMENT;
                i_start = i + 1;
            }
            else if ((line_in[i] == QUOTE)  ||
                     (line_in[i] == DOUBLE_QUOTE))
            {
                state = IN_LITERAL;
            }
            else
            {
                state = LINE_OK;
                non_ws++;
            }
            break;

        case C_COMMENT:
            if (line_in[i] == '*')
            {
                state = POST_COMMENT;
            }
            break;

        case POST_COMMENT:
            if (line_in[i] == '/')
            {
                i_end = i - 1;
                state = LINE_OK;
                non_ws = 0;
            }
            else
            {
                state = C_COMMENT;
            }
            break;

        case CPP_COMMENT:
            non_ws = 0;
            break;

        default:
            return CM_ERROR;
        }
    }

    if (state == CPP_COMMENT)
    {
        if (i > i_start)
        {
            i_end = i - 1;
        }
        state = LINE_OK;
    }
    if (state == C_COMMENT)
    {
        if (i > i_start)
        {
            i_end = i - 1;
        }
    }
    if (state == LINE_OK)
    {
        if (non_ws != 0)
        {
            i_start = i_end = 0;
        }
    }

    if (i_start > i_end)
    {
        return CM_ERROR;
    }

    *pcstart = i_start;
    *pcend   = i_end;

    return state;
}



//////////////////////////////////////////////////////////////////////////////////////////////////////////////
void                            rem_tspace                      (char *line)
{/////////////////////////////////////////////////////////////////////////////////////////////////////////////
    int                         i, n;

    if (line == NULL)
    {
        return;
    }
    n = (int)strlen (line);
    if (n == 0)
    {
        return;
    }
    n--;

    for (i = n; i > 0; i--)
    {
        if (isspace (line[i]))
        {
            line[i] = '\0';
        }
        else
        {
            break;
        }
    }
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////

int                             arg_proc                        (int                argc,
                                                                 char              *argv[],
                                                                 int               *ptsh,
                                                                 int               *pcpos,
                                                                 int               *pcsize,
                                                                 int               *pcpp,
                                                                 char             **pfn_in,
                                                                 char             **pfn_out)
{/////////////////////////////////////////////////////////////////////////////////////////////////////////////
    int                         iargs;
    int                         cpp     = 0;
    int                         tsh     = DEFAULT_TSH;
    int                         cpos    = DEFAULT_CM_START;
    int                         csize   = DEFAULT_CM_SIZE;
    char                        *fn_in  = NULL;
    char                        *fn_out = NULL;

    if (argc < 2)
    {
        printf (help_text);
        return ERROR_NO_ARGS;
    }

    for (iargs = 1; iargs < argc; iargs++)
    {
        if (_stricmp (argv[iargs], "-th") == 0)
        {
            tsh = atoi (argv[++iargs]);
            if (tsh < MIN_COL_TSH)
            {
                tsh = DEFAULT_TSH;
            }
        }
        if (_stricmp (argv[iargs], "-cpos") == 0)
        {
            cpos = atoi (argv[++iargs]);
            if (cpos < MIN_COL_TSH)
            {
                cpos = DEFAULT_CM_START;
            }
        }
        if (_stricmp (argv[iargs], "-csz") == 0)
        {
            csize = atoi (argv[++iargs]);
            if (csize < MIN_COL_TSH)
            {
                csize = DEFAULT_CM_SIZE;
            }
        }
        if (_stricmp (argv[iargs], "-cpp") == 0)
        {
            cpp = 1;
        }
        if (_stricmp (argv[iargs], "-fi") == 0)
        {
            fn_in = argv[++iargs];
        }
        if (_stricmp (argv[iargs], "-fo") == 0)
        {
            fn_out = argv[++iargs];
        }
        if (_stricmp (argv[iargs], "-help") == 0)
        {
            printf (help_text);
            return ERROR_NO_ARGS;
        }
    }

    if (tsh > cpos)
    {
        iargs   = tsh;
        tsh     = cpos;
        cpos    = iargs;
    }

    if (ptsh != NULL)
    {
        *ptsh = tsh;
    }
    if (pcpos != NULL)
    {
        *pcpos = cpos;
    }
    if (pcsize != NULL)
    {
        *pcsize = csize;
    }
    if (pcpp != NULL)
    {
        *pcpp = cpp;
    }

    if (pfn_in != NULL)
    {
        if (fn_in == NULL)
        {
            printf ("? No input file!\n");
            return ERROR_NO_INPUT;
        }
        *pfn_in = fn_in;
    }
    if (pfn_out != NULL)
    {
        if (fn_out == NULL)
        {
#ifdef  OUTPUT_MANDATORY
            printf ("? No output file!\n");
#endif // OUTPUT_MANDATORY
            return ERROR_NO_OUTPUT;
        }
        *pfn_out = fn_out;
    }
    if ((fn_in != NULL) && (fn_out != NULL))
    {
        if (strcmp (fn_in, fn_out) == 0)
        {
            printf ("? The output file cannot be the same as the input file\n");
            return ERROR_SAME_FILE;
        }
    }
    return RET_OK;
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
int                             main                            (int argc, char *argv[])
{/////////////////////////////////////////////////////////////////////////////////////////////////////////////
    char                        line[MAX_LINE_LENGTH];
    char                        cmt[MAX_LINE_LENGTH];
    int                         state   = LINE_OK;
    int                         ibeg    = 0;
    int                         iend    = 0;
    int                         cpp     = 0;
    int                         tsh     = DEFAULT_TSH;
    int                         cpos    = DEFAULT_CM_START;
    int                         csize   = DEFAULT_CM_SIZE;
    int                         cend;
    int                         ret     = 0;
    FILE                       *fp_in   = NULL;
    FILE                       *fp_out  = NULL;
    char                       *fn_in   = NULL;
    char                       *fn_out  = NULL;
    char                        fn_tmp[MAX_FN_LENGTH];
    char                        fn_bak[MAX_FN_LENGTH];

    memset (fn_tmp, 0, sizeof (fn_tmp));

    ret = arg_proc (argc, argv, &tsh, &cpos, &csize, &cpp, &fn_in, &fn_out);

    if ((ret == ERROR_NO_OUTPUT) && (fn_in != NULL))
    {
        strcpy_s (fn_tmp, MAX_FN_LENGTH, fn_in);
        strcat_s (fn_tmp, MAX_FN_LENGTH, ".tmp");
        strcpy_s (fn_bak, MAX_FN_LENGTH, fn_in);
        strcat_s (fn_bak, MAX_FN_LENGTH, ".bak");
        fn_out = fn_tmp;
        ret = RET_OK;
    }

    if (ret != RET_OK)
    {
        return ret;
    }

    fp_in = __pragma(warning(suppress:4996)) fopen (fn_in, "rt");
    if (fp_in == NULL)
    {
        printf ("No such file: %s\n", fn_in);
        return ERROR_FILE_NOT_FOUND;
    }
    fp_out = __pragma(warning(suppress:4996)) fopen (fn_out, "wt");
    if (fp_out == NULL)
    {
        printf ("Cannot create file: %s\n", fn_out);
        return ERROR_CANNOT_CREATE;
    }

    while (!feof (fp_in))
    {
        if (fgets (line, MAX_LINE_LENGTH, fp_in) == NULL)
        {
            break;
        }
        state = line_analyze (line, state, &ibeg, &iend);

        strncpy_s (cmt, MAX_LINE_LENGTH, &line[ibeg], iend - ibeg);
        rem_tspace (cmt);

        // fprintf (fp_out, "%s[%04d] cs=%4d ce=%4d: ", sz_state[state], state, ibeg, iend); // only for debug

        if ((ibeg > tsh) && (state == LINE_OK))
        {
            line[ibeg - 2] = '\0';
            rem_tspace (line);

            if (cpp)
            {
                fprintf (fp_out, "%-*s//%s\n", cpos - 1, line, cmt);
            }
            else
            {
                cend = (int)strlen (line) + 1;
                if (cend > cpos)
                {
                    cend = csize - (cend - cpos);
                }
                else
                {
                    cend = csize;
                }
                fprintf (fp_out, "%-*s/*%-*s*/\n", cpos - 1, line, cend, cmt);
            }
        }
        else
        {
            fprintf (fp_out, "%s", line);
        }
    }

    fclose (fp_in);
    fclose (fp_out);

    if (fn_tmp[0] != '\0')
    {
        if (_unlink (fn_bak) != RET_OK)
        {
            if (errno == EACCES)
            {
                printf ("? Cannot delete temp file (access)\n");
                return ERROR_DELETE;
            }
        }
        if (rename (fn_in, fn_bak) != RET_OK)
        {
            printf ("? Error renaming %s\n", fn_in);
            return ERROR_RENAME;
        }
        if (rename (fn_tmp, fn_in) != RET_OK)
        {
            printf ("? Error renaming %s\n", fn_tmp);
            return ERROR_RENAME;
        }
    }
    return RET_OK;
}
